/* rgbamerge 1/19/96 Monty */

/* When called as rgbamerge:
   specify two images or stdin+filename.  Second image is overlayed onto the
   first according to its alpha channel (the A in RGBA).  Optional x and y 
   offsets allowed.  Alpha channels properly added. */

/* When called as rgbamergeB:
   specify two images or stdin+filename.  FIRST image is overlayed onto the
   SECOND according to its alpha channel (the A in RGBA).  Optional x and y 
   offsets allowed.  Alpha channels properly added. */

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>

/* Range limiting */
inline int Range(int l, int x, int u){
  if(x>u)return(u);
  if(x>=l)return(x);
  return(l);
}

int main(int argc,char *argv[]){
  FILE *outfile=stdout;
  FILE *infile=stdin;
  FILE *infile2=stdin;
  int add=0,sub=0;
  int i=1,x,y,x2,y2,xo=0,yo=0;
  char test[8];

  /* First arg could be primary filename */

  if(argc>2 && !isdigit(argv[2][0]) && argv[2][0]!='-'){
    /* Two filenames */
    infile=fopen(argv[1],"rb");
    if(infile==NULL){
      fprintf(stderr,"rgbamerge: Unable to open first input file.\n");
      exit(-1);
    }
    infile2=fopen(argv[2],"rb");
    if(infile2==NULL){
      fprintf(stderr,"rgbamerge: Unable to open second input file.\n");
      exit(-1);
    }
    i=3;
  }else{
    if(argc>1 && !isdigit(argv[1][0]) && argv[1][0]!='-'){
      /* One filename */
      infile2=fopen(argv[1],"rb");
      if(infile2==NULL){
	fprintf(stderr,"rgbamerge: Unable to open input file.\n");
	exit(-1);
      }
      i=2;
    }else{
      fprintf(stderr,"rgbamerge: Two files (name+name or stdin+name) needed\n"
	             "           for merging process.\n");
      exit(1);
    }
  }

  if(argc>i){
    xo=atoi(argv[i]);
    if(argc>i+1)
      yo=atoi(argv[i+1]);
  }

  if(strcmp(argv[0]+strlen(argv[0])-1,"B")==0){
    FILE *temp;
    temp=infile;infile=infile2;infile2=temp;
  }
  if(strcmp(argv[0]+strlen(argv[0])-7,"rgbasub")==0 ||
     strcmp(argv[0]+strlen(argv[0])-8,"rgbasubB")==0){
    sub=-1;
  }
  if(strcmp(argv[0]+strlen(argv[0])-7,"rgbaadd")==0 ||
     strcmp(argv[0]+strlen(argv[0])-8,"rgbaaddB")==0){
    add=-1;
  }

  fscanf(infile,"%7c",test);
  test[7]='\0';
  if(strcmp(test,"RGBA32 ")){
    fprintf(stderr,"rgbamerge: background input is not an RGBA32 file.\n");
    exit(-1);
  }
  
  fscanf(infile,"%d %d%*1c",&x,&y);

  fscanf(infile2,"%7c",test);
  test[7]='\0';
  if(strcmp(test,"RGBA32 ")){
    fprintf(stderr,"rgbaquant: overlay input is not an RGBA32 file.\n");
    exit(-1);
  }
  
  fscanf(infile2,"%d %d%*1c",&x2,&y2);

  {
    int xi,yi;
    int xi2,yi2;
    int *r,*g,*b,*a;
    int *rP,*gP,*bP,*aP;
    int *r2,*g2,*b2,*a2;
    int *r2P,*g2P,*b2P,*a2P;
    int ptr;

    r=rP=malloc(x*y*sizeof(int));
    g=gP=malloc(x*y*sizeof(int));
    b=bP=malloc(x*y*sizeof(int));
    a=aP=malloc(x*y*sizeof(int));

    r2=r2P=malloc(x2*y2*sizeof(int));
    g2=g2P=malloc(x2*y2*sizeof(int));
    b2=b2P=malloc(x2*y2*sizeof(int));
    a2=a2P=malloc(x2*y2*sizeof(int));

    for(yi=0;yi<y;yi++){
      ptr=x*yi;
      for(xi=0;xi<x;xi++){
	r[ptr]=(unsigned char)fgetc(infile);
	g[ptr]=(unsigned char)fgetc(infile);
	b[ptr]=(unsigned char)fgetc(infile);
	a[ptr]=(unsigned char)fgetc(infile);
	ptr++;
      }
    }

    for(yi=0;yi<y2;yi++){
      ptr=x2*yi;
      for(xi=0;xi<x2;xi++){
	r2[ptr]=(unsigned char)fgetc(infile2);
	g2[ptr]=(unsigned char)fgetc(infile2);
	b2[ptr]=(unsigned char)fgetc(infile2);
	a2[ptr]=(unsigned char)fgetc(infile2);
	ptr++;
      }
    }

    fprintf(outfile,"RGBA32 %d %d\n",x,y);

    if(add){

      for(yi=0;yi<y;yi++){
	yi2=yi-yo;
	r=rP+yi*x;
	g=gP+yi*x;
	b=bP+yi*x;
	a=aP+yi*x;
	if(yi2>=0 && yi2<y2){
	  r2=r2P+yi2*x2;
	  g2=g2P+yi2*x2;
	  b2=b2P+yi2*x2;
	  a2=a2P+yi2*x2;
	  for(xi=0;xi<x;xi++){
	    float alpha,alpha2;
	    xi2=xi-xo;
	    alpha=a[xi]/255.;
	    alpha2=a2[xi2]/255.;
	    if(xi2>=0 && xi2<x2){
	      if(alpha>0 || alpha2>0){
		float alphaout=alpha2+(1.-alpha2)*alpha;
		r[xi]=rint((r[xi]*alpha)+
			   a2[xi2]*r2[xi2]/255.);
		g[xi]=rint((g[xi]*alpha)+
			   a2[xi2]*g2[xi2]/255.);
		b[xi]=rint((b[xi]*alpha)+
			   a2[xi2]*b2[xi2]/255.);

		a[xi]=rint(a[xi]+(255.-a[xi])*a2[xi2]/255.);
	      }
	    }
	  }
	}
      }

    }else if(sub){

      for(yi=0;yi<y;yi++){
	yi2=yi-yo;
	r=rP+yi*x;
	g=gP+yi*x;
	b=bP+yi*x;
	a=aP+yi*x;
	if(yi2>=0 && yi2<y2){
	  r2=r2P+yi2*x2;
	  g2=g2P+yi2*x2;
	  b2=b2P+yi2*x2;
	  a2=a2P+yi2*x2;
	  for(xi=0;xi<x;xi++){
	    float alpha,alpha2;
	    xi2=xi-xo;
	    alpha=a[xi]/255.;
	    alpha2=a2[xi2]/255.;
	    if(xi2>=0 && xi2<x2){
	      if(alpha>0 || alpha2>0){
		float alphaout=alpha2+(1.-alpha2)*alpha;
		r[xi]=rint((r[xi]*alpha+255*(1-alpha))*
			   (1.-alpha2+alpha2*r2[xi2]/255.));
		g[xi]=rint((g[xi]*alpha+255*(1-alpha))*
			   (1.-alpha2+alpha2*g2[xi2]/255.));
		b[xi]=rint((b[xi]*alpha+255*(1-alpha))*
			   (1.-alpha2+alpha2*b2[xi2]/255.));
		a[xi]=rint(alphaout*255.);
	      }
	    }
	  }
	}
      }

    }else{
      for(yi=0;yi<y;yi++){
	yi2=yi-yo;
	r=rP+yi*x;
	g=gP+yi*x;
	b=bP+yi*x;
	a=aP+yi*x;
	if(yi2>=0 && yi2<y2){
	  r2=r2P+yi2*x2;
	  g2=g2P+yi2*x2;
	  b2=b2P+yi2*x2;
	  a2=a2P+yi2*x2;
	  for(xi=0;xi<x;xi++){
	    xi2=xi-xo;
	    if(xi2>=0 && xi2<x2){
	      float alpha=a[xi]/255.;
	      float alpha2=a2[xi2]/255.;

	      if(alpha>0 || alpha2>0){
		float alphaout=alpha2+(1.-alpha2)*alpha;
		r[xi]=rint((r2[xi2]*alpha2+(1.-alpha2)*alpha*r[xi])/alphaout);
		g[xi]=rint((g2[xi2]*alpha2+(1.-alpha2)*alpha*g[xi])/alphaout);
		b[xi]=rint((b2[xi2]*alpha2+(1.-alpha2)*alpha*b[xi])/alphaout);
		a[xi]=rint(alphaout*255.);
	      }
	    }
	  }
	}
      }
    }

    for(yi=0;yi<y;yi++){
      r=rP+yi*x;
      g=gP+yi*x;
      b=bP+yi*x;
      a=aP+yi*x;
      for(xi=0;xi<x;xi++)
	fprintf(outfile,"%c%c%c%c",
		(unsigned char)Range(0,r[xi],255),
		(unsigned char)Range(0,g[xi],255),
		(unsigned char)Range(0,b[xi],255),
		(unsigned char)Range(0,a[xi],255));
    }
  }
  fclose(infile);
  fclose(infile2);
  fclose(outfile);
  return(0);
}    






