/* rgbaquant 1/21/97 Monty */

/* rgbaquant, rgbadither */

/* We use 3 algs; a color clusterer, a colormap area divider, and a
   diversity seeker.  Can specify the 1/2 and 2/3 splits on the
   command line if necessary, or we'll automap. */

/* First two algs based on IJG jpeg code, third on XV's 'diversity'
   algorithm */

/* RGB bits in histogram */
#define R_HIST 5
#define G_HIST 6
#define B_HIST 5

/* Distance weights */
#define R_DMUL 299
#define G_DMUL 587
#define B_DMUL 114

/* Total historgam bits; 16 bpp default */
#define HIST (1<<(R_HIST+G_HIST+B_HIST))

/* Max val of each color field in histogram key */
#define R_HISTMAX ((1<<R_HIST)-1)
#define G_HISTMAX ((1<<G_HIST)-1)
#define B_HISTMAX ((1<<B_HIST)-1)

/* How many bits to shift to 8 bits */
#define R_SHIFT (8-R_HIST)
#define G_SHIFT (8-G_HIST)
#define B_SHIFT (8-B_HIST)

/* centering weights */
#define R_ADD (1<<(R_SHIFT-1))
#define G_ADD (1<<(G_SHIFT-1))
#define B_ADD (1<<(B_SHIFT-1))

/* Range limiting */
inline int Range(int l, int x, int u){
  if(x>u)return(u);
  if(x>=l)return(x);
  return(l);
}

/* Scale to a histogram value */
#define RSCALE(x) ((int)x>>R_SHIFT)
#define GSCALE(x) ((int)x>>G_SHIFT)
#define BSCALE(x) ((int)x>>B_SHIFT)

/* Construct a histogram index from 8 bit RGB vals */
#define H24(r,g,b) ((RSCALE(r)<<(G_HIST+B_HIST))+(GSCALE(g)<<B_HIST)+BSCALE(b))

#define Hh(r,g,b) ((((int)(r))<<(G_HIST+B_HIST))+(((int)(g))<<B_HIST)+(int)(b))
     
/* Support */
#define Rh(x) ((int)x>>(G_HIST+B_HIST))
#define Gh(x) (((int)x>>B_HIST)&G_HISTMAX)
#define Bh(x) ((int)x&B_HISTMAX)
#define Rhs(x) (Rh(x)<<R_SHIFT)
#define Ghs(x) (Gh(x)<<G_SHIFT)
#define Bhs(x) (Bh(x)<<B_SHIFT)
     
/* Get an 8 bit RG or B value from a histogram index */
#define R8(x) ((Rh(x)<<R_SHIFT)+R_ADD)
#define G8(x) ((Gh(x)<<G_SHIFT)+G_ADD)
#define B8(x) ((Bh(x)<<B_SHIFT)+B_ADD)

/* distance (squared) from two RGB tuples */
#define DIST24(r1,g1,b1,r2,g2,b2) (R_DMUL*((int)r1-(int)r2)*\
				   ((int)r1-(int)r2)+\
				   G_DMUL*((int)g1-(int)g2)*\
				   ((int)g1-(int)g2)+\
				   B_DMUL*((int)b1-(int)b2)*\
				   ((int)b1-(int)b2))
     
/* This works on histogram values, not 24 bit RGB */
#define DIST(x,y) ((R_DMUL*(Rhs(x)-Rhs(y))*(Rhs(x)-Rhs(y))+\
			G_DMUL*(Ghs(x)-Ghs(y))*(Ghs(x)-Ghs(y))+\
			B_DMUL*(Bhs(x)-Bhs(y))*(Bhs(x)-Bhs(y))))

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <malloc.h>
#include <limits.h>
#include <math.h>
#include <string.h>
     
typedef struct colorcube{
  unsigned char rmin,rmax;
  unsigned char gmin,gmax;
  unsigned char bmin,bmax;
  long colorcount;
  long volume;
} colorcube;

void compute_color(colorcube *cube,long *histogram,unsigned char *rmap,
		   unsigned char *gmap,unsigned char *bmap){

  /* Mean weighted color by histogram count */
  /* Current algorithm: mean weighted by pixels (not colors) */
  /* Note it is important to get the rounding correct! */

  int r,g,b;
  int rmin,rmax,gmin,gmax,bmin,bmax;
  long total=0,count;
  long rtotal=0;
  long gtotal=0;
  long btotal=0;
  
  rmin=cube->rmin;rmax=cube->rmax;
  gmin=cube->gmin;gmax=cube->gmax;
  bmin=cube->bmin;bmax=cube->bmax;
  
  for(r=rmin;r<=rmax;r++)
    for(g=gmin;g<=gmax;g++){
      long *histptr=histogram+Hh(r,g,bmin);
      for(b=bmin;b<=bmax;b++,histptr++){
	if((count=*histptr)>0){
	  long key=histptr-histogram;
	  total+=count;
	  rtotal+=R8(key)*count;
	  gtotal+=G8(key)*count;
	  btotal+=B8(key)*count;
	}
      }
    }
  
  *rmap=Range(0,(long)rint((float)rtotal/total),255);
  *gmap=Range(0,(long)rint((float)gtotal/total),255);
  *bmap=Range(0,(long)rint((float)btotal/total),255);
}

int update_cube(colorcube *cube, long *histogram){
  int b,g,r;
  int rmin,rmax,gmin,gmax,bmin,bmax;
  long distr,distg,distb;
  long acc;
  
  rmin=cube->rmin;rmax=cube->rmax;
  gmin=cube->gmin;gmax=cube->gmax;
  bmin=cube->bmin;bmax=cube->bmax;
    
  if(rmax>rmin)
    for(r=rmin;r<=rmax;r++)
      for(g=gmin;g<=gmax;g++){
	long *histptr=histogram+Hh(r,g,bmin);
	for(b=bmin;b<=bmax;b++)
	  if(*histptr++!=0){
	    cube->rmin=rmin=r;
	    goto have_rmin;
	  }
      }
have_rmin:
  if(rmax>rmin)
    for(r=rmax;r>=rmin;r--)
      for(g=gmin;g<=gmax;g++){
	long *histptr=histogram+Hh(r,g,bmin);
	for(b=bmin;b<=bmax;b++)
	  if(*histptr++!=0){
	    cube->rmax=rmax=r;
	    goto have_rmax;
	  }
      }
have_rmax:
  if(gmax>gmin)
    for(g=gmin;g<=gmax;g++)
      for(r=rmin;r<=rmax;r++){
	long *histptr=histogram+Hh(r,g,bmin);
	for(b=bmin;b<=bmax;b++)
	  if(*histptr++!=0){
	    cube->gmin=gmin=g;
	    goto have_gmin;
	  }
      }
have_gmin:
  if(gmax>gmin)
    for(g=gmax;g>=gmin;g--)
      for(r=rmin;r<=rmax;r++){
	long *histptr=histogram+Hh(r,g,bmin);
	for(b=bmin;b<=bmax;b++)
	  if(*histptr++!=0){
	    cube->gmax=gmax=g;
	    goto have_gmax;
	  }
      }
have_gmax:
  if(bmax>bmin)
    for(b=bmin;b<=bmax;b++)
      for(r=rmin;r<=rmax;r++){
	long *histptr=histogram+Hh(r,gmin,b);
	for(g=gmin;g<=gmax;g++,histptr+=B_HIST)
	  if(*histptr!=0){
	    cube->bmin=bmin=b;
	    goto have_bmin;
	  }
      }
have_bmin:
  if(bmax>bmin)
    for(b=bmax;b>=bmin;b--)
      for(r=rmin;r<=rmax;r++){
	long *histptr=histogram+Hh(r,gmin,b);
	for(g=gmin;g<=gmax;g++,histptr+=B_HIST)
	  if(*histptr!=0){
	    cube->bmax=bmax=b;
	    goto have_bmax;
	  }
      }
have_bmax:

  distr=((rmax-rmin)<<R_SHIFT);
  distg=((gmax-gmin)<<G_SHIFT);
  distb=((bmax-bmin)<<B_SHIFT);
  cube->volume=distr*distr*R_DMUL+distg*distg*G_DMUL+distb*distb*B_DMUL;
  
  acc=0;

  for(r=rmin;r<=rmax;r++)
    for(g=gmin;g<=gmax;g++){
      long *histptr=histogram+Hh(r,g,bmin);
      for(b=bmin;b<=bmax;b++)
	if(*histptr++!=0)acc++;
    }
  cube->colorcount=acc;
  if(acc>0)return(1);
  return(0);
}

long cube_median_cut(colorcube *cube,long *histogram,long selected,
		     int split1,int split2){
  colorcube *c1,*c2;
  int r,g,b;

  while(selected<split2){

    /* Here we do the dirty work for both colors and weight; decide
       via split1 */

    c1=NULL;
    if(selected<split1){
      long i,max=0;
      for(i=0;i<selected;i++)
	if(cube[i].colorcount>max && cube[i].volume>0){
	  max=cube[i].colorcount;
	  c1=cube+i;
	}
    }else{
      long i,max=0;
      for(i=0;i<selected;i++)
	if(cube[i].volume>max && cube[i].colorcount>0){
	  max=cube[i].volume;
	  c1=cube+i;
	}
    }
    if(c1==NULL)break;		/* no splittable cubes left! */
    c2=cube+selected;
    
    /* Copy the color bounds to the new cube. */
    c2->rmax=c1->rmax; c2->gmax=c1->gmax; c2->bmax=c1->bmax;
    c2->rmin=c1->rmin; c2->gmin=c1->gmin; c2->bmin=c1->bmin;
    
    /* Choose which axis to split the cube on.*/
    r=(((int)c1->rmax-c1->rmin)<<R_SHIFT)*sqrt(R_DMUL);
    g=(((int)c1->gmax-c1->gmin)<<G_SHIFT)*sqrt(G_DMUL);
    b=(((int)c1->bmax-c1->bmin)<<B_SHIFT)*sqrt(B_DMUL);

    if(r>g && r>b)
      /* split on red */
      c2->rmin=(c1->rmax=(c1->rmax+c1->rmin)>>1)+1;
    else if(g>b)
      /* split on green */
      c2->gmin=(c1->gmax=(c1->gmax+c1->gmin)>>1)+1;
    else
      /* split on blue */
      c2->bmin=(c1->bmax=(c1->bmax+c1->bmin)>>1)+1;

    /* Update stats for cubes */
    if(update_cube(c2,histogram))
      selected++;
    if(update_cube(c1,histogram)==0){
      int num=c1-cube;
      selected--;
      memmove(cube+num,cube+num+1,(selected-num)*sizeof(cube));
    }
  }
  return selected;
}

long color_cubes(long *histogram,unsigned char *rmap,unsigned char *gmap,
		 unsigned char *bmap,int split1,int split2){
  colorcube *cube;
  int selected=1;
  int i;

  /* Based on the IJG colorcubing code */

  if(split2==0)return(0);
  cube=malloc(sizeof(colorcube)*split2);

  cube[0].rmin = 0;
  cube[0].rmax = R_HISTMAX;
  cube[0].gmin = 0;
  cube[0].gmax = G_HISTMAX;
  cube[0].bmin = 0;
  cube[0].bmax = B_HISTMAX;
  update_cube(cube,histogram);

  /* Let rip */
  selected=cube_median_cut(cube,histogram,1,split1,split2);

  /* Pick representative colors from cubes */
  for(i=0;i<selected;i++)
    compute_color(cube+i,histogram,rmap+i,gmap+i,bmap+i);

  free(cube);
  return(selected);
}

long color_diversity(long *histogram,int selected,unsigned char *rmap,
		     unsigned char *gmap,unsigned char *bmap,long quant){
      
  /* Although the code isn't from XV, the idea is.  See xvdocs.ps
     (diversity algorithm) in the XV source for the exquisite
     details. */

  /* Yeah, yeah, XV does it faster */
      
  long i;
  long lastdistance=-1;
      
  long *mindist=malloc(HIST*sizeof(long));
  long nextused=HIST;

  /* Initialize minimum distance cache */
  for(i=0;i<HIST;i++)
    mindist[i]=INT_MAX;

  /* hack the histogram such that unused cells point to the next
     used cell to speed searching */
  
  for(i=HIST-1;i>=0;i--){
    if(histogram[i]==0)histogram[i]=-(nextused-1);
    if(histogram[i]>0)nextused=i;
  }
  
  /* Normally we have colors already, but we might be running straight
     diversity.  If so, find the most popular color */

  if(selected==0){
    long max=0;
    long pick=0;
    for(i=0;i<HIST;i++)
      if(histogram[i]>max){
	max=histogram[i];
	pick=i;
      }

    rmap[0]=R8(pick);
    gmap[0]=G8(pick);
    bmap[0]=B8(pick);
    mindist[pick]=-1;
    selected++;
  }

  for(i=selected;i<quant;i++){
    int pick;
    int x,y;
    long dist;

    /* Search on max color distance */
    for(x=lastdistance+1;x<i;x++)
      for(y=0;y<HIST;y++)
	if(histogram[y]<0)
	  y=-histogram[y];
	else{
	  if(mindist[y]!=-1){
	    dist=DIST24(rmap[x],gmap[x],bmap[x],R8(y),G8(y),B8(y));
	    if(dist<mindist[y])mindist[y]=dist;
	  }
	}
    
    pick=-1;
    dist=INT_MIN;
    lastdistance=x-1;
    for(y=0;y<HIST;y++)
      if(histogram[y]<0)
	y=-histogram[y];
      else
	if(mindist[y]!=-1 && mindist[y]>dist){
	  dist=mindist[y];
	  pick=y;
	}

    if(pick!=-1){
      rmap[selected]=R8(pick);
      gmap[selected]=G8(pick);
      bmap[selected]=B8(pick);
      mindist[pick]=-1;
      selected++;
    }
  }
  free(mindist);
  return(selected);
}

int main(int argc,char *argv[]){
  FILE *outfile=stdout;
  FILE *infile=stdin;
  int i=1,x,y;
  int quant,split1,split2;
  int dither=0;
  char test[8];

  if(strcmp(argv[0]+strlen(argv[0])-10,"rgbadither")==0)dither=1;

  if(argc>1 && !isdigit(argv[i][0])){
    infile=fopen(argv[i],"rb");
    if(infile==NULL){
      fprintf(stderr,"rgbaquant: Unable to open input file.\n");
      exit(-1);
    }
    i++;
  }

  if(argc<i+1){
    fprintf(stderr,"rgbaquant: number of colors not specified.\n");
    exit(-1);
  }else{
    quant=atoi(argv[i]);
  }

  /* Set our color alg split defaults **************/

  if(dither){
    if(quant<16){
      split1=0;
      split2=0;
    }else{
      if(quant<48){
	split1=quant/2;
	split2=split1;
      }else{
	split1=(quant-(quant/3+16))>>1;
	split2=quant-(quant/3+16);
      }
    }
  }else{
    if(quant<16){
      split1=0;
      split2=0;
    }else{
	split2=quant*.60;
	split1=split2*.4;
    }
  }
  
  /*************************************************/

  if(argc>=i+2)split1=atoi(argv[++i]); 
  if(argc>=i+2)split2=atoi(argv[++i]); 

  fscanf(infile,"%7c",test);
  test[7]='\0';
  if(strcmp(test,"RGBA32 ")){
    fprintf(stderr,"rgbaquant: input is not an RGBA32 file.\n");
    exit(-1);
  }
  
  fscanf(infile,"%d %d%*1c",&x,&y);

  {
    int xi,yi;
    unsigned char *r,*g,*b,*a,rv,gv,bv;
    unsigned char *rP,*gP,*bP,*aP;
    long *histogram;
    int ptr;
    long colors=0,k;

    r=rP=malloc(x*y*sizeof(char));
    g=gP=malloc(x*y*sizeof(char));
    b=bP=malloc(x*y*sizeof(char));
    a=aP=malloc(x*y*sizeof(char));
    histogram=calloc(HIST,sizeof(long));

    for(yi=0;yi<y;yi++){
      ptr=x*yi;
      for(xi=0;xi<x;xi++){
	rv=r[ptr]=(unsigned char)fgetc(infile);
	gv=g[ptr]=(unsigned char)fgetc(infile);
	bv=b[ptr]=(unsigned char)fgetc(infile);
	a[ptr]=(unsigned char)fgetc(infile);
	k=H24(rv,gv,bv);
	if(histogram[k]==0)colors++;
	histogram[k]++;
	ptr++;
      }
    }

    if(quant>=colors)quant=colors;
    if(split2>quant)split2=quant;

    /* Do color selection */
    {
      /* List of already selected color entries */
      long selected;
      unsigned char *rmap=malloc(sizeof(unsigned char)*quant);
      unsigned char *gmap=malloc(sizeof(unsigned char)*quant);
      unsigned char *bmap=malloc(sizeof(unsigned char)*quant);

      /* OK, we need to do the IJG style color cubes... */
      selected=color_cubes(histogram,rmap,gmap,bmap,split1,split2);

      /* Now run the diversity algorithm */
      selected=color_diversity(histogram,selected,rmap,gmap,bmap,quant);

      /* Now use the histogram as a colormap */
      memset(histogram,-1,HIST*sizeof(long));

      fprintf(outfile,"RGBA32 %d %d\n",x,y);
      
      if(dither){
	
	float *rerr1=malloc(sizeof(float)*x);
	float *gerr1=malloc(sizeof(float)*x);
	float *berr1=malloc(sizeof(float)*x);
	float *rerr2=calloc(x,sizeof(float));
	float *gerr2=calloc(x,sizeof(float));
	float *berr2=calloc(x,sizeof(float));
	
	for(yi=0;yi<y;yi++){
	  float rv,gv,bv,rerr,gerr,berr;
	  float *temp;
	  
	  temp=rerr1;rerr1=rerr2;rerr2=temp;
	  temp=gerr1;gerr1=gerr2;gerr2=temp;
	  temp=berr1;berr1=berr2;berr2=temp;
	  
	  r=rP+yi*x;
	  g=gP+yi*x;
	  b=bP+yi*x;
	  a=aP+yi*x;
	  
	  if(yi&1){
	    rerr2[x-2]=0;rerr2[x-1]=0;
	    gerr2[x-2]=0;gerr2[x-1]=0;
	    berr2[x-2]=0;berr2[x-1]=0;
	    xi=x-1;
	  }else{
	    rerr2[0]=0;rerr2[1]=0;
	    gerr2[0]=0;gerr2[1]=0;
	    berr2[0]=0;berr2[1]=0;
	    xi=0;
	  }
	  
	  for(;xi>=0;){
	    int pick,histkey;
	    if(xi<0 || xi>=x)break;
	    
	    rv=Range(0.,r[xi]+rerr1[xi],255.);
	    gv=Range(0.,g[xi]+gerr1[xi],255.);
	    bv=Range(0.,b[xi]+berr1[xi],255.);
	    histkey=H24((int)rint(rv),(int)rint(gv),(int)rint(bv));
	    pick=histogram[histkey];
	    if(pick==-1){
	      /* We don't have this color yet. */
	      
	      long x;
	      long dist=INT_MAX;		    
	      for(x=0;x<quant;x++){
		long temp=DIST24(rv,gv,bv,(int)rmap[x],(int)gmap[x],
				 (int)bmap[x]);
		if(temp<dist){
		  dist=temp;
		  pick=x;
		}
	      }
	      histogram[histkey]=pick;
	    }
	    
	    rerr=rv-rmap[pick];
	    gerr=gv-gmap[pick];
	    berr=bv-bmap[pick];
	    r[xi]=rmap[pick];
	    g[xi]=gmap[pick];
	    b[xi]=bmap[pick];

	    if(yi&1){
	      if(xi>0){
		rerr1[xi-1]+=rerr*7./16.;
		gerr1[xi-1]+=gerr*7./16.;
		berr1[xi-1]+=berr*7./16.;
	      }
		
	      if(yi+1<y){
		rerr2[xi]+=rerr*5./16.;
		gerr2[xi]+=gerr*5./16.;
		berr2[xi]+=berr*5./16.;
		
		if(xi>0){
		  rerr2[xi-1]=rerr/16.;
		  gerr2[xi-1]=gerr/16.;
		  berr2[xi-1]=berr/16.;
		}
		if(xi+1<x){
		  rerr2[xi+1]+=rerr*3./16.;
		  gerr2[xi+1]+=gerr*3./16.;
		  berr2[xi+1]+=berr*3./16.;
		}
	      }
	      xi--;
	    }else{
	      if(xi+1<x){
		rerr1[xi+1]+=rerr*7./16.;
		gerr1[xi+1]+=gerr*7./16.;
		berr1[xi+1]+=berr*7./16.;
	      }
	      
	      if(yi+1<y){
		rerr2[xi]+=rerr*5./16.;
		gerr2[xi]+=gerr*5./16.;
		berr2[xi]+=berr*5./16.;
		
		if(xi>0){
		  rerr2[xi-1]+=rerr*3./16.;
		  gerr2[xi-1]+=gerr*3./16.;
		  berr2[xi-1]+=berr*3./16.;
		}
		if(xi+1<x){
		  rerr2[xi+1]=rerr/16.;
		  gerr2[xi+1]=gerr/16.;
		  berr2[xi+1]=berr/16.;
		}
	      }
	      xi++;
	    }
	  }
	}

	for(yi=0;yi<y;yi++){
	  r=rP+yi*x;
	  g=gP+yi*x;
	  b=bP+yi*x;
	  a=aP+yi*x;
	  for(xi=0;xi<x;xi++)
	    fprintf(outfile,"%c%c%c%c",r[xi],g[xi],b[xi],a[xi]);
	}

      }else{
	for(yi=0;yi<y;yi++){
	  r=rP+yi*x;
	  g=gP+yi*x;
	  b=bP+yi*x;
	  a=aP+yi*x;
	  for(xi=0;xi<x;xi++){
	    long hist=H24(r[xi],g[xi],b[xi]);
	    long pick=histogram[hist];
	    if(pick==-1){
	      /* We don't have this color yet. */
	      
	      long x;
	      long dist=INT_MAX;		    
	      for(x=0;x<quant;x++){
		long temp=DIST24(r[xi],g[xi],b[xi],(int)rmap[x],(int)gmap[x],
				 (int)bmap[x]);
		if(temp<dist){
		  dist=temp;
		  pick=x;
		}
	      }
	      histogram[hist]=pick;
	    }
	    fprintf(outfile,"%c%c%c%c",rmap[pick],gmap[pick],bmap[pick],a[xi]);
	  }
	}
      }
    }
  }
  fclose(infile);
  fclose(outfile);
  return(0);
}    
